﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class RandomSound : MVRScript
    {
		private JSONStorableStringChooser mySoundChooser;
		private JSONStorableBool myAvoidRepetition;
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableString mySoundName;
		private JSONStorableFloat mySoundProbability;

		private AudioSourceControl myAudioSource;
		private List<Sound> mySounds = new List<Sound>();
		private int myPreviousIndex1 = int.MaxValue;
		private int myPreviousIndex2 = int.MaxValue;
		private bool myNeedUIRefresh = false;
		private List<object> mySoundsUI = new List<object>();

		private class Sound
		{
			public Sound(string id, NamedAudioClip clip)
			{
				this.clip = clip;		
				this.id = id;				
				if (clip.manager == EmbeddedAudioClipManager.singleton)
					name = clip.category + "\n" + clip.displayName;
				else if (clip.manager == URLAudioClipManager.singleton)
					name = "SceneAudio\n" + clip.displayName;
				else
					name = "Unknown\n" + clip.displayName;
			}
			
			public static int SortByNameAscending(Sound a, Sound b)
			{
				return a.name.CompareTo(b.name);
			}

			public NamedAudioClip clip;
			public string name;
			public string id;
		}		

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);
						
			string storableName = string.Empty;
			if (containingAtom.type == "Person")
				storableName = "HeadAudioSource";
			else if (containingAtom.type == "AudioSource")
				storableName = "AudioSource";
			else if (containingAtom.type == "AptSpeaker")
				storableName = "AptSpeaker_Import";
			else if (containingAtom.type == "RhythmAudioSource")
				storableName = "RhythmSource";
			myAudioSource = containingAtom.GetStorableByID(storableName) as AudioSourceControl;
			
			if (myAudioSource == null)
			{
				SuperController.LogError("The RandomSound plugin needs to be placed on a Person, AudioSource, RhythmAudioSource or AptSpeaker atom.");
				Utils.SetupInfoTextNoScroll(this,
					"<color=#FF6060><size=40><b>RandomSound</b></size>\nThis plugin needs to be placed on a Person, AudioSource, RhythmAudioSource or AptSpeaker atom.</color>",
					500.0f, true
				);
				return;
			}

			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>RandomSound</b></size>\nLogicBrick that, after <i>Trigger</i> is called, selects a random <i>Sound</i> from the list to play. Sounds can be added from VaM's embedded files or from files referenced on the <i>Scene Audio</i> tab.\n\n" +
				"Use <i>Avoid Repetition</i> take the last one or two choices out of the selection, depending on the number of total available choices.\n\n" +
				"Plugin needs to be placed on a Person, AudioSource, RhythmAudioSource or AptSpeaker atom.</color>",
				500.0f, true
			);
			
			Utils.SetupButton(this, "Rescan SceneAudio", UIRefreshSounds, false);			
			mySoundChooser = new JSONStorableStringChooser("Sound", new List<string>(), "", "");
			UIDynamicPopup popup = CreateFilterablePopup(mySoundChooser, false);
			popup.popupPanelHeight = 500;
			popup.labelWidth = -8;
			
			RegisterStringChooser(mySoundChooser);
			mySoundChooser.isStorable = mySoundChooser.isRestorable = false;
			
			Utils.SetupButton(this, "Play Sound", UIPlaySound, false);
			Utils.SetupButton(this, "Stop Sound", UIStopSound, false);
			Utils.SetupButton(this, "Add Sound", UIAddSound, false);
			myAvoidRepetition = Utils.SetupToggle(this, "Avoid Repetition", true, false);
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);

			Utils.SetupButton(this, "Test Trigger", Trigger, false);
			Utils.SetupAction(this, "Trigger", Trigger);

			Utils.SetupSpacer(this, 15, false);			
			
			UIRefreshSounds();
			UIUpdate();
		}
		
		private void UIRefreshSounds()
		{
			mySounds.RemoveAll(s => s.clip.destroyed);
			
			List<string> names = new List<string>();
			List<string> uids = new List<string>();
			
			EmbeddedAudioClipManager eacm = EmbeddedAudioClipManager.singleton;			
			for (int i=0; i<eacm.embeddedClips.Length; ++i)
			{
				NamedAudioClip clip = eacm.embeddedClips[i];
				names.Add(clip.category + "\n" + clip.displayName);
				uids.Add("E#"+clip.uid);
			}
			
			URLAudioClipManager uacm = URLAudioClipManager.singleton;
			List<NamedAudioClip> urlclips = uacm.GetCategoryClips("web");
			if (urlclips != null)
			{
				for (int i=0; i<urlclips.Count; ++i)
				{
					NamedAudioClip clip = urlclips[i];
					names.Add("SceneAudio\n" + clip.displayName);
					uids.Add("U#"+clip.uid);
				}
			}
			
			mySoundChooser.choices = uids;
			mySoundChooser.displayChoices = names;
			
			if (uids.Count == 0)
				mySoundChooser.val = "";
			else if (!uids.Contains(mySoundChooser.val))
				mySoundChooser.val = uids[0];
			
			UIUpdate();
		}
		

		private void UIUpdate()
		{
			Utils.RemoveUIElements(this, mySoundsUI);
			
			if (mySounds.Count > 0)
			{
				UIDynamicTextInfo header = Utils.SetupInfoOneLine(this, "<size=30><b>Sounds</b></size>", false);
				mySoundsUI.Add(header);
				for (int i=0; i<mySounds.Count; ++i)
				{
					Sound sound = mySounds[i];
					UIDynamicLabelXButton label = Utils.SetupLabelXButton(this, sound.name, () => UIRemoveSound(sound), false);
					mySoundsUI.Add(label);
				}
			}
		}
		
		private NamedAudioClip UIGetClip(string id)
		{
			if (id.StartsWith("E#"))
				return EmbeddedAudioClipManager.singleton.GetClip(id.Substring(2));
			else if (id.StartsWith("U#"))
				return URLAudioClipManager.singleton.GetClip(id.Substring(2));
			return null;
		}
		
		private void UIPlaySound()
		{
			string id = mySoundChooser.val;
			NamedAudioClip clip = UIGetClip(id);
			if (clip != null && myAudioSource != null)
				myAudioSource.PlayNowClearQueue(clip);
		}
		
		private void UIStopSound()
		{
			if (myAudioSource != null)
				myAudioSource.StopAndClearQueue();
		}

		private void UIAddSound()
		{
			string id = mySoundChooser.val;
			NamedAudioClip clip = UIGetClip(id);
			if (clip != null)
			{
				mySounds.Add(new Sound(id, clip));
				mySounds.Sort(Sound.SortByNameAscending);
				UIUpdate();
			}
		}

		private void UIRemoveSound(Sound sound)
		{
			mySounds.Remove(sound);
			UIUpdate();
		}
		
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray slist = new JSONArray();
				for (int i=0; i<mySounds.Count; ++i)
					slist.Add("", mySounds[i].id);
				jc["Sounds"] = slist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				mySounds.Clear();
 				JSONArray slist = jc["Sounds"].AsArray;
				for (int i=0; i<slist.Count; ++i)
				{
					string id = slist[i].Value;
					NamedAudioClip clip = UIGetClip(id);
					if (clip != null)
						mySounds.Add(new Sound(id, clip));
				}
				mySounds.Sort(Sound.SortByNameAscending);
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				UIUpdate();
			}
		}

		private void OnDestroy()
		{
			mySounds.Clear();
			Utils.OnDestroyUI();
		}

		private void Trigger()
		{
			int count = mySounds.Count;
			if (count == 0 || !myAcceptTrigger.val)
				return;

			if (count < 3 || myPreviousIndex1 >= count)
				myPreviousIndex1 = int.MaxValue;
			if (count < 4 || myPreviousIndex2 >= count)
				myPreviousIndex2 = int.MaxValue;
			if (myPreviousIndex1 == myPreviousIndex2)
				myPreviousIndex2 = int.MaxValue;
			if (myAvoidRepetition.val && myPreviousIndex1 < int.MaxValue)
				--count;
			if (myAvoidRepetition.val && myPreviousIndex2 < int.MaxValue)
				--count;
						
			int idx = UnityEngine.Random.Range(0, count);			
			if (idx >= Mathf.Min(myPreviousIndex1, myPreviousIndex2))
				++idx;
			if (idx >= Mathf.Max(myPreviousIndex1, myPreviousIndex2))
				++idx;
					
			myPreviousIndex2 = myPreviousIndex1;
			myPreviousIndex1 = idx;

			if (myAudioSource != null)
				myAudioSource.PlayNowClearQueue(mySounds[idx].clip);
		}
    }
}
